#pragma once

#include <iostream>
#include <numeric>
#include <algorithm>
#include <cstring>
#include <stdio.h>
#include <stdlib.h>
#include <error.h>
#include <assert.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <vector>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <map>
#include <set>
#include <omp.h>
#ifdef USE_MPI
#include <mpi.h>
#endif

#define FNAME_LEN 256

#if defined(CLOCK_MONOTONIC)
#define CLOCK CLOCK_MONOTONIC
#elif defined(CLOCK_REALTIME)
#define CLOCK CLOCK_REALTIME
#else
#error "Failed to find a timing clock."
#endif

typedef uint32_t vertex_id_t;
typedef uint64_t edge_id_t;
typedef double weight_t;

using namespace std;


typedef map<vertex_id_t, weight_t> map_edges_t;
typedef map<vertex_id_t, map_edges_t> map_graph_t;

/* The graph data structure */
struct graph_t
{
    /**********************************************************
     The minimal graph repesentation consists of:
     n        -- the number of vertices
     m        -- the number of edges
     endV     -- an array of size m that stores the 
                 destination ID of an edge <src->dest>.
     rowsIndices -- an array of size n+1 that stores the degree 
                 (out-degree in case of directed graphs) and pointers to
                 the endV array. The degree of vertex i is given by 
                 rowsIndices[i+1]-rowsIndices[i], and the edges out of i are
                 stored in the contiguous block endV[rowsIndices[i] .. rowsIndices[i+1]-1].
     Vertices are ordered from 0 in our internal representation
    ***********************************************************/
    vertex_id_t n;
    edge_id_t m;
    edge_id_t *rowsIndices;
    vertex_id_t *endV;
    
    /* Edge weights */
    weight_t *weights;
    weight_t min_weight, max_weight;
    
    /* other graph parameters */
    int scale; /* log2 of vertices number */
    double avg_vertex_degree; /* relation m / n */
    bool directed;
    char filename[FNAME_LEN]; /* filename for output graph */

    /* Distributed version variables */
    int nproc, rank;
    vertex_id_t local_n; /* local vertices number */
    edge_id_t local_m; /* local edges number */
    edge_id_t* num_edges_of_any_process;
};

/* write graph to file */
void writeBinaryGraph(graph_t *G, char *filename);

/* read graph from file */
void readGraph(graph_t *G, char *filename);

/* free graph memory */
void freeGraph(graph_t *G);

/* main */
void* build_graph_structure(graph_t &graph);
void community_detection(size_t n, void *graph, vector<vertex_id_t> &component_id);
void free_graph_structure(void *graph);

/* main_mpi */
void gen_SSCA2_graph_MPI(graph_t *G);
void gen_RMAT_graph_MPI(graph_t *G);
void gen_LFR_graph_MPI(graph_t *G);
void community_detection(vertex_id_t n, vertex_id_t local_n, void* graph, vector<vertex_id_t> &component_id);

struct LFR_params{
    double mu;
    double ctau;
    double dtau;
    vertex_id_t maxd;
    vertex_id_t maxs;
    vertex_id_t mind;
    vertex_id_t mins;
};


void readGraph_singleFile_MPI(graph_t *G, char *filename);

/* returns global number of the edge from local number */
edge_id_t edge_to_global(edge_id_t, graph_t *);

#define MOD_SIZE(v) ((v) % size)
#define DIV_SIZE(v) ((v) / size)
#define MUL_SIZE(x) ((x) * size)

/* returns number of vertex owner, v - the global vertex number, TotVertices - the global number of vertices, size - the number of processes */
inline int VERTEX_OWNER(const vertex_id_t v, const vertex_id_t TotVertices, const int size)
{
    vertex_id_t mod_size = MOD_SIZE(TotVertices);
    vertex_id_t div_size = DIV_SIZE(TotVertices);
    if (!mod_size) {
        return v / div_size;
    } else {
        if (v / (div_size + 1) < mod_size) {
            return v / (div_size + 1);
        } else {
            return (v - mod_size * (div_size + 1)) / div_size + mod_size;
        }
    }
}

/* returns local vertex number, v - the global vertex number, TotVertices - the global number of vertices, size - the number of processes, rank - the process number */
inline vertex_id_t VERTEX_LOCAL(const vertex_id_t v, const vertex_id_t TotVertices, const int size, const int rank) 
{
    if (MOD_SIZE(TotVertices) <= (unsigned int)rank) {
        return ((v - MOD_SIZE(TotVertices) * (DIV_SIZE(TotVertices) + 1)) % DIV_SIZE(TotVertices));
    } else {
        return (v % (DIV_SIZE(TotVertices) + 1));
    }
}

/* returns global vertex number, v_local - the local vertex number, TotVertices - the global number of vertices, size - the number of processes, rank - the process number */
inline vertex_id_t VERTEX_TO_GLOBAL(const vertex_id_t v_local, const vertex_id_t TotVertices, const int size, const int rank)
{
    if (MOD_SIZE(TotVertices) > (unsigned int)rank) {
        return ((DIV_SIZE(TotVertices) + 1) * rank + (vertex_id_t)v_local);
    } else {
        return (MOD_SIZE(TotVertices) * (DIV_SIZE(TotVertices) + 1) + DIV_SIZE(TotVertices) * (rank - MOD_SIZE(TotVertices)) + v_local);
    }
}

inline vertex_id_t get_local_n(graph_t *G)
{
    vertex_id_t TotVertices = G->n;
    int size = G->nproc;
    int rank = G->rank;
    vertex_id_t mod_size = MOD_SIZE(TotVertices);
    vertex_id_t div_size = DIV_SIZE(TotVertices);
    if (!mod_size) {
        return div_size;
    } else {
        if (rank < (long long)mod_size)
            return div_size + 1;
        else
            return div_size;
    }
}

